from __future__ import annotations

from cleo.io.inputs.argv_input import ArgvInput
from cleo.io.inputs.token_parser import TokenParser


class StringInput(ArgvInput):
    """
    Represents an input provided as a string
    """

    def __init__(self, input: str) -> None:
        super().__init__([])

        self._set_tokens(self._tokenize(input))

    def _tokenize(self, input: str) -> list[str]:
        return TokenParser().parse(input)



from __future__ import annotations

from cleo._utils import find_similar_names


class CleoError(Exception):
    """
    Base Cleo exception.
    """

    exit_code: int | None = None


class CleoLogicError(CleoError):
    """
    Raised when there is error in command arguments
    and/or options configuration logic.
    """


class CleoRuntimeError(CleoError):
    """
    Raised when command is called with invalid options or arguments.
    """


class CleoValueError(CleoError):
    """
    Raised when wrong value was given to Cleo components.
    """


class CleoNoSuchOptionError(CleoError):
    """
    Raised when command does not have given option.
    """


class CleoUserError(CleoError):
    """
    Base exception for user errors.
    """


class CleoMissingArgumentsError(CleoUserError):
    """
    Raised when called command was not given required arguments.
    """


class CleoCommandNotFoundError(CleoUserError):
    """
    Raised when called command does not exist.
    """

    def __init__(self, name: str, commands: list[str] | None = None) -> None:
        message = f'The command "{name}" does not exist.'

        if commands:
            suggested_names = find_similar_names(name, commands)

            if suggested_names:
                if len(suggested_names) == 1:
                    message += "\n\nDid you mean this?\n    "
                else:
                    message += "\n\nDid you mean one of these?\n    "

                message += "\n    ".join(suggested_names)

        super().__init__(message)


class CleoNamespaceNotFoundError(CleoUserError):
    """
    Raised when called namespace has no commands.
    """

    def __init__(self, name: str, namespaces: list[str] | None = None) -> None:
        message = f'There are no commands in the "{name}" namespace.'

        if namespaces:
            suggested_names = find_similar_names(name, namespaces)

            if suggested_names:
                if len(suggested_names) == 1:
                    message += "\n\nDid you mean this?\n    "
                else:
                    message += "\n\nDid you mean one of these?\n    "

                message += "\n    ".join(suggested_names)

        super().__init__(message)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from cleo.application import Application
from cleo.commands.command import Command
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.definition import Definition
from cleo.io.inputs.option import Option
from cleo.io.outputs.output import Type


if TYPE_CHECKING:
    from cleo.io.io import IO


class Descriptor:
    def describe(self, io: IO, obj: Any, **options: Any) -> None:
        self._io = io

        if isinstance(obj, Argument):
            self._describe_argument(obj, **options)
        elif isinstance(obj, Option):
            self._describe_option(obj, **options)
        elif isinstance(obj, Definition):
            self._describe_definition(obj, **options)
        elif isinstance(obj, Command):
            self._describe_command(obj, **options)
        elif isinstance(obj, Application):
            self._describe_application(obj, **options)

    def _write(self, content: str, decorated: bool = True) -> None:
        self._io.write(
            content, new_line=False, type=Type.NORMAL if decorated else Type.RAW
        )

    def _describe_argument(self, argument: Argument, **options: Any) -> None:
        raise NotImplementedError()

    def _describe_option(self, option: Option, **options: Any) -> None:
        raise NotImplementedError()

    def _describe_definition(self, definition: Definition, **options: Any) -> None:
        raise NotImplementedError()

    def _describe_command(self, command: Command, **options: Any) -> None:
        raise NotImplementedError()

    def _describe_application(self, application: Application, **options: Any) -> None:
        raise NotImplementedError()



from __future__ import annotations

from io import StringIO
from typing import TYPE_CHECKING

from cleo.io.buffered_io import BufferedIO
from cleo.io.inputs.argv_input import ArgvInput
from cleo.io.inputs.string_input import StringInput
from cleo.io.outputs.buffered_output import BufferedOutput


if TYPE_CHECKING:
    from cleo.commands.command import Command
    from cleo.io.outputs.output import Verbosity


class CommandTester:
    """
    Eases the testing of console commands.
    """

    def __init__(self, command: Command) -> None:
        self._command = command
        self._io = BufferedIO()
        self._inputs: list[str] = []
        self._status_code: int | None = None

    @property
    def command(self) -> Command:
        return self._command

    @property
    def io(self) -> BufferedIO:
        return self._io

    @property
    def status_code(self) -> int | None:
        return self._status_code

    def execute(
        self,
        args: str = "",
        inputs: str | None = None,
        interactive: bool | None = None,
        verbosity: Verbosity | None = None,
        decorated: bool | None = None,
        supports_utf8: bool = True,
    ) -> int:
        """
        Executes the command
        """
        application = self._command.application

        input: StringInput | ArgvInput = StringInput(args)
        if (
            application is not None
            and application.definition.has_argument("command")
            and self._command.name is not None
        ):
            name = self._command.name
            if " " in name:
                # If the command is namespaced we rearrange
                # the input to parse it as a single argument
                argv = [application.name, self._command.name] + input._tokens

                input = ArgvInput(argv)
            else:
                input = StringInput(name + " " + args)

        self._io.set_input(input)
        assert isinstance(self._io.output, BufferedOutput)
        assert isinstance(self._io.error_output, BufferedOutput)
        self._io.output.set_supports_utf8(supports_utf8)
        self._io.error_output.set_supports_utf8(supports_utf8)

        if inputs is not None:
            self._io.input.set_stream(StringIO(inputs))

        if interactive is not None:
            self._io.interactive(interactive)

        if verbosity is not None:
            self._io.set_verbosity(verbosity)

        if decorated is not None:
            self._io.decorated(decorated)

        self._status_code = self._command.run(self._io)

        return self._status_code



from __future__ import annotations

from typing import Callable

from cleo.commands.command import Command
from cleo.exceptions import CleoCommandNotFoundError
from cleo.loaders.command_loader import CommandLoader


Factory = Callable[[], Command]


class FactoryCommandLoader(CommandLoader):
    """
    A simple command loader using factories to instantiate commands lazily.
    """

    def __init__(self, factories: dict[str, Factory]) -> None:
        self._factories = factories

    @property
    def names(self) -> list[str]:
        return list(self._factories.keys())

    def has(self, name: str) -> bool:
        return name in self._factories

    def get(self, name: str) -> Command:
        if name not in self._factories:
            raise CleoCommandNotFoundError(name)

        factory = self._factories[name]

        return factory()



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from cleo.ui.table_cell_style import TableCellStyle


class TableCell(str):
    def __new__(
        cls,
        value: str = "",
        rowspan: int = 1,
        colspan: int = 1,
        style: TableCellStyle | None = None,
    ) -> TableCell:
        return super().__new__(cls, value)

    def __init__(
        self,
        value: str = "",
        rowspan: int = 1,
        colspan: int = 1,
        style: TableCellStyle | None = None,
    ) -> None:
        self._rowspan = rowspan
        self._colspan = colspan
        self._style = style

    @property
    def rowspan(self) -> int:
        return self._rowspan

    @property
    def colspan(self) -> int:
        return self._colspan

    @property
    def style(self) -> TableCellStyle | None:
        return self._style



from __future__ import annotations

from typing import Any

from cleo.io.inputs.argument import Argument
from cleo.io.inputs.option import Option


def argument(
    name: str,
    description: str | None = None,
    optional: bool = False,
    multiple: bool = False,
    default: Any | None = None,
) -> Argument:
    return Argument(
        name,
        required=not optional,
        is_list=multiple,
        description=description,
        default=default,
    )


def option(
    long_name: str,
    short_name: str | None = None,
    description: str | None = None,
    flag: bool = True,
    value_required: bool = True,
    multiple: bool = False,
    default: Any | None = None,
) -> Option:
    return Option(
        long_name,
        short_name,
        flag=flag,
        requires_value=value_required,
        is_list=multiple,
        description=description,
        default=default,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from cleo.events.event import Event


if TYPE_CHECKING:
    from cleo.commands.command import Command
    from cleo.io.io import IO


class ConsoleEvent(Event):
    """
    An event that gives access to the IO of a command.
    """

    def __init__(self, command: Command, io: IO) -> None:
        super().__init__()

        self._command = command
        self._io = io

    @property
    def command(self) -> Command:
        return self._command

    @property
    def io(self) -> IO:
        return self._io



from __future__ import annotations

import sys

from typing import TYPE_CHECKING
from typing import TextIO

from cleo.io.io import IO


if TYPE_CHECKING:
    from cleo.io.outputs.output import Output


class Cursor:
    def __init__(self, io: IO | Output, input: TextIO | None = None) -> None:
        if isinstance(io, IO):
            io = io.output

        self._output = io

        if input is None:
            input = sys.stdin

        self._input = input

    def move_up(self, lines: int = 1) -> Cursor:
        self._output.write(f"\x1b[{lines}A")

        return self

    def move_down(self, lines: int = 1) -> Cursor:
        self._output.write(f"\x1b[{lines}B")

        return self

    def move_right(self, columns: int = 1) -> Cursor:
        self._output.write(f"\x1b[{columns}C")

        return self

    def move_left(self, columns: int = 1) -> Cursor:
        self._output.write(f"\x1b[{columns}D")

        return self

    def move_to_column(self, column: int) -> Cursor:
        self._output.write(f"\x1b[{column}G")

        return self

    def move_to_position(self, column: int, row: int) -> Cursor:
        self._output.write(f"\x1b[{row + 1};{column}H")

        return self

    def save_position(self) -> Cursor:
        self._output.write("\x1b7")

        return self

    def restore_position(self) -> Cursor:
        self._output.write("\x1b8")

        return self

    def hide(self) -> Cursor:
        self._output.write("\x1b[?25l")

        return self

    def show(self) -> Cursor:
        self._output.write("\x1b[?25h\x1b[?0c")

        return self

    def clear_line(self) -> Cursor:
        """
        Clears all the output from the current line.
        """
        self._output.write("\x1b[2K")

        return self

    def clear_line_after(self) -> Cursor:
        """
        Clears all the output from the current line after the current position.
        """
        self._output.write("\x1b[K")

        return self

    def clear_output(self) -> Cursor:
        """
        Clears all the output from the cursors' current position
        to the end of the screen.
        """
        self._output.write("\x1b[0J")

        return self

    def clear_screen(self) -> Cursor:
        """
        Clears the entire screen.
        """
        self._output.write("\x1b[2J")

        return self



from __future__ import annotations

import sys

from typing import TYPE_CHECKING
from typing import Any
from typing import Sequence

from cleo.exceptions import CleoLogicError
from cleo.io.inputs.option import Option


if TYPE_CHECKING:
    from cleo.io.inputs.argument import Argument


class Definition:
    """
    A Definition represents a set of command line arguments and options.
    """

    def __init__(self, definition: Sequence[Argument | Option] | None = None) -> None:
        self._arguments: dict[str, Argument] = {}
        self._required_count = 0
        self._has_a_list_argument = False
        self._has_optional = False
        self._options: dict[str, Option] = {}
        self._shortcuts: dict[str, str] = {}

        if definition is None:
            definition = []

        self.set_definition(definition)

    @property
    def arguments(self) -> list[Argument]:
        return list(self._arguments.values())

    @property
    def argument_count(self) -> int:
        if self._has_a_list_argument:
            return sys.maxsize

        return len(self._arguments)

    @property
    def required_argument_count(self) -> int:
        return self._required_count

    @property
    def argument_defaults(self) -> dict[str, Any]:
        values = {}

        for argument in self._arguments.values():
            values[argument.name] = argument.default

        return values

    @property
    def options(self) -> list[Option]:
        return list(self._options.values())

    @property
    def option_defaults(self) -> dict[str, Any]:
        values = {}
        for option in self._options.values():
            values[option.name] = option.default

        return values

    def set_definition(self, definition: Sequence[Argument | Option]) -> None:
        arguments = []
        options = []

        for item in definition:
            if isinstance(item, Option):
                options.append(item)
            else:
                arguments.append(item)

        self.set_arguments(arguments)
        self.set_options(options)

    def set_arguments(self, arguments: list[Argument]) -> None:
        self._arguments = {}
        self._required_count = 0
        self._has_a_list_argument = False
        self._has_optional = False
        self.add_arguments(arguments)

    def add_arguments(self, arguments: list[Argument]) -> None:
        for argument in arguments:
            self.add_argument(argument)

    def add_argument(self, argument: Argument) -> None:
        if argument.name in self._arguments:
            raise CleoLogicError(
                f'An argument with name "{argument.name}" already exists'
            )

        if self._has_a_list_argument:
            raise CleoLogicError("Cannot add an argument after a list argument")

        if argument.is_required() and self._has_optional:
            raise CleoLogicError("Cannot add a required argument after an optional one")

        if argument.is_list():
            self._has_a_list_argument = True

        if argument.is_required():
            self._required_count += 1
        else:
            self._has_optional = True

        self._arguments[argument.name] = argument

    def argument(self, name: str | int) -> Argument:
        if not self.has_argument(name):
            raise ValueError(f'The "{name}" argument does not exist')

        if isinstance(name, int):
            arguments = list(self._arguments.values())
            return arguments[name]

        return self._arguments[name]

    def has_argument(self, name: str | int) -> bool:
        arguments: dict[str, Argument] | list[Argument]
        if isinstance(name, int):
            arguments = list(self._arguments.values())
        else:
            arguments = self._arguments

        try:
            arguments[name]  # type: ignore[index]
        except (KeyError, IndexError):
            return False

        return True

    def set_options(self, options: list[Option]) -> None:
        self._options = {}
        self._shortcuts = {}
        self.add_options(options)

    def add_options(self, options: list[Option]) -> None:
        for option in options:
            self.add_option(option)

    def add_option(self, option: Option) -> None:
        if option.name in self._options and option != self._options[option.name]:
            raise CleoLogicError(f'An option named "{option.name}" already exists')

        if option.shortcut:
            for shortcut in option.shortcut.split("|"):
                if (
                    shortcut in self._shortcuts
                    and option.name != self._shortcuts[shortcut]
                ):
                    raise CleoLogicError(
                        f'An option with shortcut "{shortcut}" already exists'
                    )

        self._options[option.name] = option

        if option.shortcut:
            for shortcut in option.shortcut.split("|"):
                self._shortcuts[shortcut] = option.name

    def option(self, name: str) -> Option:
        if not self.has_option(name):
            raise ValueError(f'The option "--{name}" option does not exist')

        return self._options[name]

    def has_option(self, name: str) -> bool:
        return name in self._options

    def has_shortcut(self, shortcut: str) -> bool:
        return shortcut in self._shortcuts

    def option_for_shortcut(self, shortcut: str) -> Option:
        return self._options[self.shortcut_to_name(shortcut)]

    def shortcut_to_name(self, shortcut: str) -> str:
        if shortcut not in self._shortcuts:
            raise ValueError(f'The "-{shortcut}" option does not exist')

        return self._shortcuts[shortcut]

    def synopsis(self, short: bool = False) -> str:
        elements = []

        if short and self._options:
            elements.append("[options]")
        elif not short:
            for option in self._options.values():
                value = ""
                if option.accepts_value():
                    formatted = (
                        option.name.upper()
                        if option.requires_value()
                        else f"[{option.name.upper()}]"
                    )
                    value = f" {formatted}"

                shortcut = ""
                if option.shortcut:
                    shortcut = f"-{option.shortcut}|"

                elements.append(f"[{shortcut}--{option.name}{value}]")

        if elements and self._arguments:
            elements.append("[--]")

        tail = ""
        for argument in self._arguments.values():
            element = f"<{argument.name}>"
            if argument.is_list():
                element += "..."

            if not argument.is_required():
                element = "[" + element
                tail += "]"

            elements.append(element)

        return " ".join(elements) + tail



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from cleo.exceptions import CleoCommandNotFoundError


if TYPE_CHECKING:
    from cleo.application import Application
    from cleo.commands.command import Command


class ApplicationDescription:

    GLOBAL_NAMESPACE = "_global"

    def __init__(
        self,
        application: Application,
        namespace: str | None = None,
        show_hidden: bool = False,
    ) -> None:
        self._application: Application = application
        self._namespace = namespace
        self._show_hidden = show_hidden
        self._namespaces: dict[str, dict[str, str | list[str]]] = {}
        self._commands: dict[str, Command] = {}
        self._aliases: dict[str, Command] = {}

        self._inspect_application()

    @property
    def namespaces(self) -> dict[str, dict[str, str | list[str]]]:
        return self._namespaces

    @property
    def commands(self) -> dict[str, Command]:
        return self._commands

    def command(self, name: str) -> Command:
        if name in self._commands:
            return self._commands[name]
        if name in self._aliases:
            return self._aliases[name]
        raise CleoCommandNotFoundError(name)

    def _inspect_application(self) -> None:
        namespace = None
        if self._namespace:
            namespace = self._application.find_namespace(self._namespace)

        all_commands = self._application.all(namespace)

        for namespace, commands in self._sort_commands(all_commands):
            names = []

            for name, command in commands:
                if not command.name or command.hidden:
                    continue

                if command.name == name:
                    self._commands[name] = command
                else:
                    self._aliases[name] = command

                names.append(name)

            self._namespaces[namespace] = {"id": namespace, "commands": names}

    def _sort_commands(
        self, commands: dict[str, Command]
    ) -> list[tuple[str, list[tuple[str, Command]]]]:
        """
        Sorts command in alphabetical order
        """
        namespaced_commands: dict[str, dict[str, Command]] = defaultdict(dict)
        for name, command in commands.items():
            key = self._application.extract_namespace(name, 1) or "_global"
            namespaced_commands[key][name] = command

        namespaced_commands_lst: dict[str, list[tuple[str, Command]]] = {}
        for namespace, commands in namespaced_commands.items():
            namespaced_commands_lst[namespace] = sorted(
                commands.items(), key=lambda x: x[0]
            )

        return sorted(namespaced_commands_lst.items(), key=lambda x: x[0])
